import numpy as np
import unittest
import Fred.backend as fred

class TestContinuousFrechet(unittest.TestCase):

    def test_zigzag1d(self):
        a = fred.Curve([0.0, 1.0, 0.0, 1.0])
        b = fred.Curve([0.0, 0.75, 0.25, 1.0])
        c = fred.Curve([0.0, 1.0])
        self.assertEqual(round(fred.continuous_frechet(a, b).value, 2), 0.25)
        self.assertEqual(round(fred.continuous_frechet(a, c).value, 1), 0.5)
        
    def test_longsegment(self):
        a = fred.Curve([0.0,500.0e3, 1.0e6])
        b = fred.Curve([0.0, 1.0e6])
        self.assertEqual(round(fred.continuous_frechet(a, b).value, 3), 0.0)

class TestDiscreteFrechet(unittest.TestCase):
    
    def test_zigzag1d(self):
        a = fred.Curve([0.0, 1.0, 0.0, 1.0])
        b = fred.Curve([0.0, 0.75, 0.25, 1.0])
        c = fred.Curve([0.0, 1.0])
        self.assertEqual(fred.discrete_frechet(a, b).value, 0.25)
        self.assertEqual(fred.discrete_frechet(a, c).value, 1.0)
        
    def test_longsegment(self):
        a = fred.Curve([0.0,500.0e3, 1.0e6])
        b = fred.Curve([0.0, 1.0e6])
        self.assertEqual(fred.discrete_frechet(a, b).value, 500000.0)
        
class TestDiscreteDynamicTimeWarping(unittest.TestCase):
    
    def test_zigzag1d(self):
        a = fred.Curve([0.0, 1.0, 0.0, 1.0])
        b = fred.Curve([0.0, 0.75, 0.25, 1.0])
        c = fred.Curve([0.0, 1.0])
        self.assertEqual(fred.discrete_dynamic_time_warping(a, b).value, 0.5)
        self.assertEqual(fred.discrete_dynamic_time_warping(a, c).value, 1.0)
        
    def test_longsegment(self):
        a = fred.Curve([0.0,500.0e3, 1.0e6])
        b = fred.Curve([0.0, 1.0e6])
        self.assertEqual(fred.discrete_dynamic_time_warping(a, b).value, 500000.0)

if __name__ == '__main__':
    unittest.main()



from .backend import *

config = Config()

def plot_curve(*curves, vertex_markings=True, savename=None, saveextension=None):
    import matplotlib.pyplot as plt
    from mpl_toolkits.mplot3d import Axes3D
    max_compl = 1
    max_dim = 1
    fig = plt.figure()
    ax = None
    if max_dim >= 3:
        ax = fig.gca(projection='3d')
    else:
        ax = fig.gca()
    for curve in curves:
        if isinstance(curve, backend.Curve):
            max_compl = max(max_compl, curve.complexity)
            max_dim = max(max_dim, curve.dimensions)
        elif isinstance(curve, backend.Curves):
            for curv in curve:
                max_compl = max(max_compl, curv.complexity)
                max_dim = max(max_dim, curv.dimensions)
        elif isinstance(curve, backend.Clustering_Result):
            for curv in curve:
                max_compl = max(max_compl, curv.complexity)
                max_dim = max(max_dim, curv.dimensions)
    for curve in curves:
        if isinstance(curve, backend.Curve):
            if curve.dimensions >= 3:
                p = ax.plot(curve.values[:, 0], curve.values[:, 1], curve.values[:, 2], linestyle='--', marker='o', label = curve.name, markersize = 7, markevery = curve.complexity)
                if vertex_markings:
                    ax.plot(curve.values[1:, 0], curve.values[1:, 1], linestyle='', marker='x', label = None, color = p[0].get_color(), markersize = 7)
            elif curve.dimensions == 2:
                p = ax.plot(curve.values[:, 0], curve.values[:, 1], linestyle='--', marker='o', label = curve.name, markersize = 7, markevery = curve.complexity)
                if vertex_markings:
                    ax.plot(curve.values[1:, 0], curve.values[1:, 1], linestyle='', marker='x', label = None, color = p[0].get_color(), markersize = 7)
            else:
                p = ax.plot([i * max_compl / len(curve) for i in range(len(curve))], curve.values, linestyle='--', marker='o', label = curve.name, markersize = 7, markevery = curve.complexity)
                if vertex_markings:
                    ax.plot([i * max_compl / len(curve) for i in range(1, len(curve))], curve.values[1:], linestyle='', marker='x', label = None, color = p[0].get_color(), markersize = 7)
        elif isinstance(curve, backend.Curves):
            for curv in curve:
                if curv.dimensions >= 3:
                    p = ax.plot(curv.values[:, 0], curv.values[:, 1], curv.values[:, 2], linestyle='--', marker='o', label = curv.name, markersize = 7, markevery = curv.complexity)
                    if vertex_markings:
                        ax.plot(curv.values[1:, 0], curv.values[1:, 1], linestyle='', marker='x', label = None, color = p[0].get_color(), markersize = 7)
                elif curv.dimensions == 2:
                    p = plt.plot(curv.values[:, 0], curv.values[:, 1], linestyle='--', marker='o', label = curv.name, markersize = 7, markevery = curv.complexity)
                    if vertex_markings:
                        plt.plot(curv.values[1:, 0], curv.values[1:, 1], linestyle='', marker='x', label = None, color = p[0].get_color(), markersize = 7)
                else:
                    p = plt.plot([i * max_compl / len(curv) for i in range(len(curv))], curv.values, linestyle='--', marker='o', label = curv.name, markersize = 7, markevery = curv.complexity)
                    if vertex_markings:
                        plt.plot([i * max_compl / len(curv) for i in range(1, len(curv))], curv.values[1:], linestyle='', marker='x', label = None, color = p[0].get_color(), markersize = 7)
        elif isinstance(curve, backend.Clustering_Result):
            for curv in curve:
                if curv.dimensions >= 3:
                    p = ax.plot(curv.values[:, 0], curv.values[:, 1], curv.values[:, 2], linestyle='-', marker='o', label = curv.name, markersize = 7, markevery = curv.complexity)
                    if vertex_markings:
                        ax.plot(curv.values[1:, 0], curv.values[1:, 1], linestyle='', marker='x', label = None, color = p[0].get_color(), markersize = 7)
                elif curv.dimensions == 2:
                    p = plt.plot(curv.values[:, 0], curv.values[:, 1], linestyle='-', marker='o', label = curv.name, markersize = 7, markevery = curv.complexity)
                    if vertex_markings:
                        plt.plot(curv.values[1:, 0], curv.values[1:, 1], linestyle='', marker='x', label = None, color = p[0].get_color(), markersize = 7)
                else:
                    p = plt.plot([i * max_compl / len(curv) for i in range(len(curv))], curv.values, linestyle='-', marker='o', label = curv.name, markersize = 7, markevery = curv.complexity)
                    if vertex_markings:
                        plt.plot([i * max_compl / len(curv) for i in range(1, len(curv))], curv.values[1:], linestyle='', marker='x', label = None, color = p[0].get_color(), markersize = 7)
    ax.legend(title='Curve names:')
    ax.set_title('Fred Curves')
    if savename is None:
        plt.show()
    else:
        plt.savefig("{}.{}".format(savename, saveextension), dpi=150)
    plt.close()
    
def plot_clustering(clustering_result, curves, vertex_markings=True, savename=None, saveextension=None):
    if not (isinstance(clustering_result, backend.Clustering_Result) and isinstance(curves, backend.Curves)):
        print("Check parameters!")
        return
    if len(clustering_result.assignment) < 1:
        print("compute_assignment was not called! calling now")
        clustering_result.compute_assignment(curves)
    import matplotlib.pyplot as plt
    from mpl_toolkits.mplot3d import Axes3D
    import matplotlib.colors as mcolors
    colors = list(mcolors.BASE_COLORS)
    if len(clustering_result) > len(colors):
        colors = list(mcolors.TABLEAU_COLORS)
    if len(clustering_result) > len(colors):
        colors = list(mcolors.mcolors.CSS4_COLORS)
    max_compl = 1
    max_dim = 1
    fig = plt.figure()
    ax = None
    if max_dim >= 3:
        ax = fig.gca(projection='3d')
    else:
        ax = fig.gca()
    for curve in curves:
        max_compl = max(max_compl, curve.complexity)
        max_dim = max(max_dim, curve.dimensions)
    for i, curve in enumerate(clustering_result):
        if curve.dimensions >= 3:
            p = ax.plot(curve.values[:, 0], curve.values[:, 1], curve.values[:, 2], linestyle='-', marker='o', label = curve.name, color = colors[i], markersize = 7, markevery = curve.complexity)
            if vertex_markings:
                ax.plot(curve.values[1:, 0], curve.values[1:, 1], linestyle='', marker='x', label = None, color = p[0].get_color(), markersize = 7)
        elif curve.dimensions == 2:
            p = ax.plot(curve.values[:, 0], curve.values[:, 1], linestyle='-', marker='o', label = curve.name, color = colors[i], markersize = 7, markevery = curve.complexity)
            if vertex_markings:
                ax.plot(curve.values[1:, 0], curve.values[1:, 1], linestyle='', marker='x', label = None, color = p[0].get_color(), markersize = 7)
        else:
            p = ax.plot([i * max_compl / len(curve) for i in range(len(curve))], curve.values, linestyle='-', marker='o', label = curve.name, color = colors[i], markersize = 7, markevery = curve.complexity)
            if vertex_markings:
                ax.plot([i * max_compl / len(curve) for i in range(1, len(curve))], curve.values[1:], linestyle='', marker='x', label = None, color = p[0].get_color(), markersize = 7)
    for i in range(len(clustering_result.assignment)):
        for j in range(clustering_result.assignment.count(i)):
            curve = curves[clustering_result.assignment.get(i,j)]
            if curve.dimensions >= 3:
                p = ax.plot(curve.values[:, 0], curve.values[:, 1], curve.values[:, 2], linestyle='--', marker='o', label = curve.name, color = colors[i], markersize = 7, markevery = curve.complexity)
                if vertex_markings:
                    ax.plot(curve.values[1:, 0], curve.values[1:, 1], linestyle='', marker='x', label = None, color = p[0].get_color(), markersize = 7)
            elif curve.dimensions == 2:
                p = ax.plot(curve.values[:, 0], curve.values[:, 1], linestyle='--', marker='o', label = curve.name, color = colors[i], markersize = 7, markevery = curve.complexity)
                if vertex_markings:
                    ax.plot(curve.values[1:, 0], curve.values[1:, 1], linestyle='', marker='x', label = None, color = p[0].get_color(), markersize = 7)
            else:
                p = ax.plot([i * max_compl / len(curve) for i in range(len(curve))], curve.values, linestyle='--', marker='o', label = curve.name, color = colors[i], markersize = 7, markevery = curve.complexity)
                if vertex_markings:
                    ax.plot([i * max_compl / len(curve) for i in range(1, len(curve))], curve.values[1:], linestyle='', marker='x', label = None, color = p[0].get_color(), markersize = 7)
    ax.legend(title='Curve names:')
    ax.set_title('Fred Curves')
    if savename is None:
        plt.show()
    else:
        plt.savefig("{}.{}".format(savename, saveextension), dpi=150)
    plt.close()



from . import backend

from cvxopt import matrix, solvers
solvers.options['show_progress'] = False
import numpy as np

def curve_within_radii(T, radii):
    l = len(T)
    start_point = _intersection_point(T, radii, [1] * l, [0] * l)
    if not start_point[0]:
        return (False, )
    end_point = _intersection_point(T, radii, [T[i].points - 1 for i in range(0, l)], [1] * l)
    if not end_point[0]:
        return (False, )
    borders = [([1] * len(T), start_point[1]), ([T[i].points for i in range(0, l)], end_point[1])]
    _compute_borders(T, radii, np.array([1] * l, dtype=np.uint64), borders)
    borders = sorted(borders, key=lambda border: border[0])
    solution = list()
    solution.append(borders[-1])
    alpha = [T[i].points for i in range(0, l)]
    for i in range(len(borders)-2, -1, -1):
        diff = np.array(alpha) - np.array(borders[i][0])
        bools = any(diff > 1) or any(np.array(borders[i][0]) > np.array(alpha)) or all(np.array(borders[i][0]) == np.array(alpha))
        if not bools:
            solution.append(borders[i])
            alpha = borders[i][0]
    if solution[-1] != borders[0]:
        return (False, )
    solution = reversed(solution)
    return (True, backend.Curve(np.array([sol[1] for sol in solution], dtype="double")))
    
def _compute_borders(T, radii, alpha, borders):
    for j in range(0, len(T)):
        if alpha[j] + 1 < T[j].points:
            border = _cellborder(T, radii, alpha, j, 1)
            if border[0]:
                mask = np.zeros((len(T), ))
                mask[j] = 1
                borders.append(([alpha[i] + border[1][0][i] for i in range(0, len(alpha))], border[1][1]))
                _compute_borders(T, radii, alpha + mask, borders)

def _intersection_point(T, radii, alpha, values):
    lines = list()
    for i in range(0, len(T)):
        lines.append([(1-values[i])*T[i][alpha[i]-1][k] + values[i]*T[i][alpha[i]][k] for k in range(0, T[i].dimensions)])
    
    c = matrix([0.] * T[0].dimensions)
    G = []
    for i in range(0, T[0].dimensions):
        m = len(T) * ([0.] + i * [0.] + [-1] + (T[0].dimensions - i - 1) * [0.])
        G.append(m)
        
    G = matrix(G)
    h = []
    for i in range(0, len(T)):
        h += [radii[i]] + [-lines[i][j] for j in range(0, T[i].dimensions)]
    h = matrix(h)
    dims = {'l': 0, 'q': [1+T[0].dimensions] * len(T), 's': [0]}
    sol = solvers.conelp(c, G, h, dims)
    return sol['status'] == 'optimal', np.array([s for s in sol['x']]) if sol['x'] is not None else None

def _cellborder(T, radii, alpha, j, value):
    lines = list()
    for i in range(0, len(T)):
        if (i == j): 
            lines.append(([(1-value)*T[i][int(alpha[i]-1)][k] + value*T[i][int(alpha[i])][k] for k in range(0, T[i].dimensions)],))
        else:
            lines.append(([T[i][int(alpha[i]-1)][k] - T[i][int(alpha[i])][k] for k in range(0, T[i].dimensions)], 
                          [T[i][int(alpha[i]-1)][k] for k in range(0, T[i].dimensions)]))
        
    dims = {'l': 2 * (len(T)-1), 'q': [1+T[0].dimensions] * len(T), 's': [0]}

    c = matrix([0.] * (len(T) - 1) + [0.] * T[0].dimensions)
    G = list()
    
    for i in range(0, len(T)):
        if i == j:
            pass
        elif i < j:
            m = i * [0.] + [1.] + (len(T) - i - 2) * [0.]
            m += i * [0.] + [-1.] + (len(T) - i - 2) * [0.]
            m += (i * (T[i].dimensions + 1)) * [0.] + [0.] + [-lines[i][0][k] for k in range(0, T[i].dimensions)] + ((len(T) - i - 1) * (T[i].dimensions + 1)) * [0.]
            G.append(m)
        else:
            m = (i - 1) * [0.] + [1.] + (len(T) - i - 1) * [0.]
            m += (i - 1) * [0.] + [-1.] + (len(T) - i - 1) * [0.]
            m += (i * (T[i].dimensions + 1)) * [0.] + [0.] + [-lines[i][0][k] for k in range(0, T[i].dimensions)] + ((len(T) - i - 1) * (T[i].dimensions + 1)) * [0.]
            G.append(m)
        
    for i in range(0, T[0].dimensions):
        m = (2 * (len(T) - 1)) * [0.]
        m += len(T) * ([0.] + i * [0.] + [-1] + (T[0].dimensions - i - 1) * [0.])
        G.append(m)
        
    G = matrix(G)
    
    h = [1.] * (len(T) - 1)
    h += [0.] * (len(T) - 1)
    for i in range(0, len(T)):
        h += [radii[i]]
        if i == j:
            h += [-lines[i][0][k] for k in range(0, T[i].dimensions)]
        else:
            h += [-lines[i][1][k] for k in range(0, T[i].dimensions)]
    h = matrix(h)
    try:
        sol = solvers.conelp(c, G, h, dims)
    except Exception as e:
        print(e)
    if sol['x'] is not None:
        x = sol['x']
        y = np.empty((len(T),))
        for i in range(0, len(T)):
            if i < j:
                y[i] = x[i]
            elif i == j:
                y[j] = value
            else:
                y[i] = x[i-1]
    return sol['status'] == 'optimal', (y, np.array([s for s in sol['x'][len(T)-1:]])) if sol['x'] is not None else None
